# -*- coding: utf-8 -*-
"""PyNCM 网易云音乐 Python API / 下载工具

PyNCM 包装的网易云音乐 API 的使用非常简单::
    
    >>> from pyncm import apis
    # 登录
    >>> apis.LoginViaCellphone(phone="[..]", password="[..]", ctcode=86, remeberLogin=True)
    # 获取歌曲信息    
    >>> apis.track.GetTrackAudio(29732235)
    {'data': [{'id': 29732235, 'url': 'http://m701.music...
    # 获取歌曲详情
    >>> apis.track.GetTrackDetail(29732235)    
    {'songs': [{'name': 'Supernova', 'id': 2...
    # 获取歌曲评论
    >>> apis.track.GetTrackComments(29732235)    
    {'isMusician': False, 'userId': -1, 'topComments': [], 'moreHot': True, 'hotComments': [{'user': {'locationInfo': None, 'liveIn ...

PyNCM 的所有 API 请求都将经过单例的 `pyncm.Session` 发出，管理此单例可以使用::

    >>> session = pyncm.GetCurrentSession()
    >>> pyncm.SetCurrentSession(session)
    >>> pyncm.SetNewSession()

PyNCM 同时提供了相应的 Session 序列化函数，用于其储存及管理::

    >>> save = pyncm.DumpSessionAsString()
    >>> pyncm.SetNewSession(
            pyncm.LoadSessionFromString(save)
        )

# 注意事项
    - (PR#11) 海外用户可能经历 460 "Cheating" 问题，可通过添加以下 Header 解决: `X-Real-IP = 118.88.88.88`    
"""
from typing import Text, Union
from time import time
from .utils.crypto import EapiEncrypt, EapiDecrypt, HexCompose
import requests, logging, json
logger = logging.getLogger("pyncm.api")

__version__ = "1.6.8.2"

DEVICE_ID_DEFAULT = "pyncm!"
# This sometimes fails with some strings, for no particular reason. Though `pyncm!` seem to work everytime..?
# Though with this, all pyncm users would then be sharing the same device Id.
# Don't think that would be of any issue though...
"""默认 deviceID"""

class Session(requests.Session):
    """# Session
    实现网易云音乐登录态 / API请求管理

    - HTTP方面，`Session`的配置方法和 `requests.Session` 完全一致，如配置 Headers::

    GetCurrentSession().headers['X-Real-IP'] = '1.1.1.1'

    - 该 Session 其他参数也可被修改::

    GetCurrentSession().force_http = True # 优先 HTTP

    获取其他具体信息请参考该文档注释
    """    
    HOST = "music.163.com"
    """网易云音乐 API 服务器域名，可直接改为代理服务器之域名"""
    UA_DEFAULT = "Mozilla/5.0 (linux@github.com/mos9527/pyncm) Chrome/PyNCM.%s" % __version__        
    """Weapi 使用的 UA"""
    UA_EAPI = "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Safari/537.36 Chrome/91.0.4472.164 NeteaseMusicDesktop/2.10.2.200154"
    """EAPI 使用的 UA，不推荐更改"""
    UA_LINUX_API = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36"
    """曾经的 Linux 客户端 UA，不推荐更改"""
    force_http = False
    """优先使用 HTTP 作 API 请求协议"""
    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": self.UA_DEFAULT,
            "Referer": self.HOST,
        }
        self.login_info = {"success": False, "tick": time(), "content": None}
        self.eapi_config = {
            "os": "pc",
            "appver": "",
            "osver": "",
            "deviceId": DEVICE_ID_DEFAULT,
        }        
        self.csrf_token = ""

    # region Shorthands
    @property
    def deviceId(self):
        """设备 ID"""
        return self.eapi_config["deviceId"]
    @deviceId.setter
    def deviceId(self,v):
        self.eapi_config["deviceId"] = str(v)       
 
    @property
    def uid(self):
        """用户 ID"""
        return self.login_info["content"]["account"]["id"] if self.logged_in else 0

    @property
    def nickname(self):
        """登陆用户的昵称"""
        return (
            self.login_info["content"]["profile"]["nickname"] if self.logged_in else ""
        )

    @property
    def lastIP(self):
        """登陆时，上一次登陆的 IP"""
        return (
            self.login_info["content"]["profile"]["lastLoginIP"]
            if self.logged_in
            else ""
        )

    @property
    def vipType(self):
        """账号 VIP 等级"""
        return self.login_info["content"]["profile"]["vipType"] if self.logged_in else 0

    @property
    def logged_in(self):
        """是否已经登陆"""
        return self.login_info["success"]

    # endregion
    def request(
        self, method: str, url: Union[str, bytes, Text], *a, **k
    ) -> requests.Response:
        """发起 HTTP(S) 请求
        该函数与 `requests.Session.request` 有以下不同：
        - 使用 SSL 与否取决于 `force_http`
        - 不强调协议（只用 HTTP(S)），不带协议的链接会自动补上 HTTP(S)
        
        Args:
            method (str): HTTP Verb
            url (Union[str, bytes, Text]): Complete/Partial HTTP URL

        Returns:
            requests.Response
        """
        if url[:4] != "http":
            url = "https://%s%s" % (self.HOST, url)
        if self.force_http:
            url = url.replace("https:", "http:")
        return super().request(method, url, *a, **k)

    # region symbols for loading/reloading authentication info
    _session_info = {
        "eapi_config":(
            lambda self: getattr(self, "eapi_config"),
            lambda self, v: setattr(self, "eapi_config", v),
        ),
        "login_info": (
            lambda self: getattr(self, "login_info"),
            lambda self, v: setattr(self, "login_info", v),
        ),
        "csrf_token": (
            lambda self: getattr(self, "csrf_token"),
            lambda self, v: setattr(self, "csrf_token", v),
        ),
        "cookies": (
            lambda self: [
                {"name": c.name, "value": c.value, "domain": c.domain, "path": c.path}
                for c in getattr(self, "cookies")
            ],
            lambda self, cookies: [
                getattr(self, "cookies").set(**cookie) for cookie in cookies
            ],
        ),
    }

    def dump(self) -> dict:
        """以 `dict` 导出登录态"""
        return {
            name: self._session_info[name][0](self)
            for name in self._session_info.keys()
        }

    def load(self, dumped):
        """从 `dict` 加载登录态"""
        for k, v in dumped.items():
            self._session_info[k][1](self, v)
        return True

    # endregion


class SessionManager:
    """PyNCM Session 单例储存对象"""
    def __init__(self) -> None:
        self.session = Session()

    def get(self):
        return self.session

    def set(self, session):
        self.session = session

    # region Session serialization
    @staticmethod
    def stringify_legacy(session: Session) -> str:
        """（旧）序列化 `Session` 为 `str`"""
        return EapiEncrypt("pyncm", json.dumps(session.dump()))["params"]

    @staticmethod
    def parse_legacy(dump: str) -> Session:
        """（旧）反序列化 `str` 为 `Session`"""
        session = Session()
        dump = HexCompose(dump)
        dump = EapiDecrypt(dump).decode()
        dump = dump.split("-36cd479b6b5-")
        assert dump[0] == "pyncm"  # check magic
        session.load(json.loads(dump[1]))  # loading config dict
        return session

    @staticmethod
    def stringify(session: Session) -> str:
        """序列化 `Session` 为 `str`"""
        from json import dumps
        from zlib import compress
        from base64 import b64encode
        return 'PYNCM' + b64encode(compress(dumps(session.dump()).encode())).decode()

    @staticmethod
    def parse(dump : str) -> Session:
        """反序列化 `str` 为 `Session`"""
        if dump[:5] == 'PYNCM': # New marshaler (compressed,base64 encoded) has magic header
            from json import loads
            from zlib import decompress
            from base64 import b64decode
            session = Session()
            session.load(loads(decompress(b64decode(dump[5:])).decode()))
            return session
        else:
            return SessionManager.parse_legacy(dump)

    # endregion

sessionManager = SessionManager()

def GetCurrentSession() -> Session:
    """获取当前正在被 PyNCM 使用的 Session / 登录态"""
    return sessionManager.get()


def SetCurrentSession(session: Session):
    """设置当前正在被 PyNCM 使用的 Session / 登录态"""
    sessionManager.set(session)


def SetNewSession():
    """设置新的被 PyNCM 使用的 Session / 登录态"""
    sessionManager.set(Session())


def LoadSessionFromString(dump: str) -> Session:
    """从 `str` 加载 Session / 登录态"""
    session = SessionManager.parse(dump)
    return session


def DumpSessionAsString(session: Session) -> str:
    """从 Session / 登录态 导出 `str`"""
    return SessionManager.stringify(session)



# -*- coding: utf-8 -*-
"""Sports FM - 跑步FM APIs"""
from .. import EapiCryptoRequest
import json


@EapiCryptoRequest
def GetSportsFMRecommendations(limit=3, bpm: int = 50, e_r=True):
    """移动端 - 小程序 - 获取跑步FM歌曲推荐

    Args:
        limit (int, optional): 获取数目. Defaults to 3.
        bpm (int): 实为‘每分钟步数’. Defaults to 50.
        e_r (bool, optional): 未知. Defaults to True.

    Returns:
        dict
    """
    return "/eapi/radio/sport/get", {
        "limit": str(limit),
        "bpm": str(bpm),
        "e_r": str(e_r).lower(),
    }


@EapiCryptoRequest
def GetCalculatedSportsFMStatus(
    distance=0, maxbpm=0, time=0, songList=[], steps=0, bpm=0, e_r=True
):
    """移动端 - 小程序 - 计算运动详情

    Args:
        distance (int, optional): 距离 - 单位未知. Defaults to 0.
        maxbpm (int, optional): 最大心率 - 暂存. Defaults to 0.
        time (int, optional): 运动用时 - 秒为单位. Defaults to 0.
        songList (list, optional): 播放过的歌曲列表. Defaults to [].
        steps (int, optional): 行进的步数. Defaults to 0.
        bpm (int, optional): 结束时刻心率. Defaults to 0.
        e_r (bool, optional): 未知. Defaults to True.

    Returns:
        dict
    """
    return "/eapi/radio/sport/calculate", {
        "distance": str(distance),
        "maxbpm": str(maxbpm),
        "time": str(time),
        "songList": json.dumps(songList),
        "steps": str(steps),
        "bpm": str(bpm),
        "e_r": str(e_r).lower(),
    }



"""二维码登录 Demo"""
from __init__ import assert_dep

assert_dep("qrcode")
import time, qrcode


def login():
    from pyncm.apis.login import (
        GetCurrentLoginStatus,
        WriteLoginInfo,
        LoginQrcodeUnikey,
        LoginQrcodeCheck,
    )

    uuid = LoginQrcodeUnikey()["unikey"]  # 获取 UUID
    print("UUID", uuid)

    url = f"https://music.163.com/login?codekey={uuid}"  # 二维码内容即这样的 URL
    img = qrcode.make(url)
    input("按 Enter 以显示二维码,扫描完毕请关闭")
    img.show()
    while True:
        for dot in ["...", ".. ", ".  "]:

            rsp = LoginQrcodeCheck(uuid)  # 检测扫描状态
            print(f"{rsp['code']} -- {rsp['message']}", dot, end="\r")
            if rsp["code"] == 803 or rsp["code"] == 800:
                # 登录成功
                print(f"{rsp['code']} -- {rsp['message']}", "...")
                WriteLoginInfo(GetCurrentLoginStatus())
                return True
            time.sleep(1)


if __name__ == "__main__":
    print("二维码登录测试")
    assert login(), "登陆失败"



from pyncm.apis.login import LoginViaCellphone,SetSendRegisterVerifcationCodeViaCellphone,GetRegisterVerifcationStatusViaCellphone
from pyncm import GetCurrentSession,DumpSessionAsString, LoadSessionFromString, SetCurrentSession
from pprint import pprint

def login():
    import inquirer
    query = inquirer.prompt(
        [
            inquirer.Text("phone", message="手机号"),            
            inquirer.Text("ctcode", message="国家代码(默认86)"),
        ]
    )
    phone,ctcode = query["phone"], query["ctcode"] or 86
    if inquirer.confirm('使用手机验证码登陆？'):
        result = SetSendRegisterVerifcationCodeViaCellphone(phone,ctcode)        
        if not result.get('code',0) == 200:
            pprint(result)
        else:
            print('[-] 已发送验证码')    
        while True:
            captcha = inquirer.text("输入验证码")
            verified = GetRegisterVerifcationStatusViaCellphone(phone,captcha,ctcode)
            pprint(verified)
            if verified.get('code',0) == 200:
                print('[-] 验证成功')
                break
        result = LoginViaCellphone(phone,captcha=captcha,ctcode=ctcode)
        pprint(result)
    else:
        password = inquirer.password("输入密码")
        result = LoginViaCellphone(phone,password=password,ctcode=ctcode)
        pprint(result)
    print('[!] 登录态 Session:',DumpSessionAsString(GetCurrentSession()))
    print('[-] 此后可通过 SetCurrentSession(LoadSessionFromString("PYNCMe...")) 恢复当前登录态')
    return True

if __name__ == "__main__":
    print("[-] 登录测试")
    assert login(), "登陆失败"



# -*- coding: utf-8 -*-
"""PyNCM 网易云音乐 API 封装

本模块提供各种网易云音乐 API 的 Python 封装，及编写新 API 所需的加密修饰器

模块使用可参考子模块说明，如::

    ...
    def GetTrackAudio(song_ids: list, bitrate=320000, encodeType="aac"):
        PC 端 - 获取歌曲音频详情（文件URL、MD5...）

        Args:
            song_ids (list): 歌曲 ID
            bitrate (int, optional): 比特率 (96k SQ 320k HQ 320k+ Lossless/SQ). Defaults to 320000
            encodeType (str, optional) Defaults to `aac`
    ...

使用 `pyncm.apis.GetTrackAudio` 即可

编写新 API
- EAPI
    在获得相关 API 路径和参数后(可参考 [Binaryify 大佬的视频 ](https://www.bilibili.com/video/BV1PY41147r9))

    可参考已有 API 函数格式编写相关 PyNCM API. 各API模式修饰器已在本文件注释
- WEAPI
    由于Weapi使用不对称加密，截取Weapi可从浏览器内断点尝试

    一般在 `encSecKey` 处下断点即可，具体方法不再阐述
"""

from random import randrange
from functools import wraps
from requests.models import Response
from ..utils.crypto import (
    EapiDecrypt,
    EapiEncrypt,
    LinuxApiEncrypt,
    WeapiEncrypt,
    AbroadDecrypt,
)
from .. import GetCurrentSession, logger
import json, urllib.parse

class LoginRequiredException(Exception):
    pass

class LoginFailedException(Exception):
    pass

LOGIN_REQUIRED = LoginRequiredException("需要登录")

def _BaseWrapper(requestFunc):
    """API加密函数通用修饰器
    
    实际使用请参考以下其他 Wrapper::
        
        LoginRequiredApi
        UserIDBasedApi
        WeapiCryptoRequest
        LapiCryptoRequest
        EapiCryptoRequest
    """
    @wraps(requestFunc)
    def apiWrapper(apiFunc):
        @wraps(apiFunc)
        def wrapper(*a, **k):
            ret = apiFunc(*a, **k)
            url, payload = ret[:2]
            method = ret[-1] if ret[-1] in ["POST", "GET"] else "POST"            
            logger.debug('API=%s %s url=%s deviceId=%s payload=%s' % (
                requestFunc.__name__.split('Crypto')[0].upper(),
                method,
                url,
                GetCurrentSession().deviceId,
                payload)
            )
            rsp = requestFunc(url, payload, method)
            try:
                payload = rsp.text if isinstance(rsp, Response) else rsp
                payload = payload.decode() if not isinstance(payload, str) else payload
                payload = json.loads(payload.strip("\x10"))
                if "abroad" in payload and payload["abroad"]: 
                    # Addresses Issue #15
                    # This however, has some unforeseen side-effects. Mainly due to its diffrences
                    # with non-abraod responses.                    
                    logger.warn('Detected "abroad" payload. API response might differ in format!')
                    real_payload = AbroadDecrypt(payload["result"])
                    payload = {"result":json.loads(real_payload)}
                    # We'll let the user know about it
                    payload['abroad'] = True
                return payload
            except json.JSONDecodeError as e:
                logger.error('Response is not valid JSON : %s' % e)
                logger.error('Response : %s',rsp)
                return rsp

        return wrapper

    return apiWrapper


def LoginRequiredApi(func):
    """API 需要事先登录"""
    @wraps(func)
    def wrapper(*a, **k):
        if not GetCurrentSession().login_info["success"]:
            raise LOGIN_REQUIRED
        return func(*a, **k)

    return wrapper


def UserIDBasedApi(func):
    """API 第一参数为用户 ID，而该参数可留 0 而指代已登录的用户 ID"""
    @wraps(func)
    def wrapper(user_id=0, *a, **k):
        if user_id == 0 and GetCurrentSession().login_info["success"]:
            user_id = GetCurrentSession().uid
        elif user_id == 0:
            raise LOGIN_REQUIRED
        return func(user_id, *a, **k)

    return wrapper


def EapiEncipered(func):
    """函数值有 Eapi 加密 - 解密并返回原文"""
    @wraps(func)
    def wrapper(*a, **k):
        payload = func(*a, **k)
        try:
            return EapiDecrypt(payload).decode()
        except:
            return payload

    return wrapper

@_BaseWrapper
def WeapiCryptoRequest(url, plain, method):
    """Weapi - 适用于 网页端、小程序、手机端部分 APIs"""
    payload = json.dumps({**plain, "csrf_token": GetCurrentSession().csrf_token})
    return GetCurrentSession().request(
        method,
        url.replace("/api/", "/weapi/"),
        params={"csrf_token": GetCurrentSession().csrf_token},
        data={**WeapiEncrypt(payload)},
    )

# 来自 https://github.com/Binaryify/NeteaseCloudMusicApi
@_BaseWrapper
def LapiCryptoRequest(url, plain, method):
    """Linux API - 适用于Linux客户端部分APIs"""
    payload = {"method": method, "url": GetCurrentSession().HOST + url, "params": plain}
    payload = json.dumps(payload)
    return GetCurrentSession().request(
        method,
        "/api/linux/forward",
        headers={"User-Agent": GetCurrentSession().UA_LINUX_API},
        data={**LinuxApiEncrypt(payload)},
    )

# 来自 https://github.com/Binaryify/NeteaseCloudMusicApi
@_BaseWrapper
@EapiEncipered
def EapiCryptoRequest(url, plain, method):
    """Eapi - 适用于新版客户端绝大部分API"""    
    payload = {**plain, "header": json.dumps({
        **GetCurrentSession().eapi_config,
        "requestId": str(randrange(20000000,30000000))
    })}
    digest = EapiEncrypt(urllib.parse.urlparse(url).path.replace("/eapi/", "/api/"), json.dumps(payload))    
    request = GetCurrentSession().request(
        method,
        url,
        headers={"User-Agent": GetCurrentSession().UA_EAPI, "Referer": None},
        cookies={
            **GetCurrentSession().eapi_config
        },
        data={
            **digest
        },
    )
    return request.content

from . import (
    miniprograms,
    album,
    cloud,
    cloudsearch,
    login,
    playlist,
    track,
    user,
    video,
)



# -*- coding: utf-8 -*-
"""网易云搜索 - Cloudsearch APIs"""
from . import WeapiCryptoRequest

SONG = 1
"""歌曲"""
ALBUM = 10
"""专辑"""
ARTIST = 100
"""艺术家"""
PLAYLIST = 1000
"""歌单®"""
USER = 1002
"""用户"""
MV = 1004
"""MV"""
LYRICS = 1006
"""歌词"""
DJ = 1009
"""电台"""
VIDEO = 1014
"""视频"""


@WeapiCryptoRequest
def GetSearchResult(keyword: str, stype=SONG, limit=30, offset=0):
    """网页端 - 搜索某类型关键字

    Args:
        keyword (str): 搜索关键字
        type ([int, optional): 搜索类型 (cloudsearch.SONG/...). Defaults to SONG
        limit (int, optional): 单次获取量. Defaults to 30.
        offset (int, optional): 获取偏移数. Defaults to 0.

    Returns:
        dict
    """
    return "/weapi/cloudsearch/get/web", {
        "s": str(keyword),
        "type": str(stype),
        "limit": str(limit),
        "offset": str(offset),
    }



# -*- coding: utf-8 -*-
"""Essential implementations of some of netease's security algorithms"""
import base64
from . import HexCompose,HexDigest,HashHexDigest,RandomString,security
from .aes import AES
# region secrets
WEAPI_RSA_PUBKEY = (
    int("00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7",16),
    int("10001",16) # textbook rsa without padding
)
# AES keys & IVs targets AES-128 mode
WEAPI_AES_KEY    = "0CoJUm6Qyw8W8jud" # cbc
WEAPI_AES_IV     = "0102030405060708" # cbc
LINUXAPI_AES_KEY = "rFgB&h#%2?^eDg:Q" # ecb
EAPI_DIGEST_SALT = "nobody%(url)suse%(text)smd5forencrypt"
EAPI_DATA_SALT   = "%(url)s-36cd479b6b5-%(text)s-36cd479b6b5-%(digest)s"
EAPI_AES_KEY     = "e82ckenh8dichen8" # ecb
# endregion

# region Cryptographic algorithims
def PKCS7_pad(data,bs=AES.BLOCKSIZE):
    return data + (bs - len(data) % bs) * chr(bs - len(data) % bs)

def PKCS7_unpad(data,bs=AES.BLOCKSIZE):
        pad = data[-1]
        if not pad in range(0,bs):
            return data # hack : data isn't padded
        return data[:-pad]

def AESEncrypt(data:str, key:str, iv='',mode=AES.MODE_CBC):    
    cipher = AES(key.encode())        
    if mode == AES.MODE_CBC:
        return cipher.encrypt_cbc_nopadding(PKCS7_pad(data).encode(),iv.encode())
    else:
        return cipher.encrypt_ecb_nopadding(PKCS7_pad(data).encode())

def AESDecrypt(data:str,key:str,iv='',mode=AES.MODE_CBC):       
    cipher = AES(key.encode())
    if type(data) == str : data = data.enc
    if mode == AES.MODE_CBC:
        return PKCS7_unpad(cipher.decrypt_cbc_nopadding(data,iv.encode()))
    else:
        return PKCS7_unpad(cipher.decrypt_ecb_nopadding(data))

def RSAEncrypt(data:str, n , e , reverse=True):
    m = data if not reverse else reversed(data)        
    m, e, N = int(''.join(m).encode('utf-8').hex(), 16), e , n        
    r = pow(m,e,N)
    return HexCompose(hex(r)[2:].zfill(256))
# endregion

# region api-specific crypto routines
def WeapiEncrypt(params, aes_key2=None):
    """Implements /weapi/ Asymmetric encryption"""
    aes_key2 = aes_key2 or RandomString(16)
    params = str(params)
    # 1st go,encrypt the text with aes_key and aes_iv
    params = str(base64.encodebytes(AESEncrypt(
        data=params, key=WEAPI_AES_KEY, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
    # 2nd go,encrypt the ENCRYPTED text again,with the 2nd key and aes_iv
    params = str(base64.encodebytes(AESEncrypt(
        data=params, key=aes_key2, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
    # 3rd go,generate RSA encrypted encSecKey
    encSecKey = HexDigest(RSAEncrypt(aes_key2, *WEAPI_RSA_PUBKEY))        
    return {
        'params': params,
        'encSecKey': encSecKey
    }

def AbroadDecrypt(result):
    """Decrypts 'abroad:True' messages"""
    return security.c_decrypt_abroad_message(result)

def EapiEncrypt(url,params):
    """Implements EAPI request encryption"""
    url,params = str(url),str(params)
    digest = HashHexDigest(EAPI_DIGEST_SALT % {'url':url,'text':params})
    params = EAPI_DATA_SALT % ({'url':url,'text':params,'digest':digest})
    return {
        'params':HexDigest(AESEncrypt(params,key=EAPI_AES_KEY,mode=AES.MODE_ECB))
    }
    
def EapiDecrypt(cipher):
    """Implements EAPI response decryption"""
    cipher = bytearray(cipher) if isinstance(cipher,str) else cipher
    return AESDecrypt(cipher,EAPI_AES_KEY,mode=AES.MODE_ECB) if cipher else cipher

def LinuxApiEncrypt(params):
    """Implements Linux/Deepin client API encryption"""
    params = str(params)
    return {
        'eparams':HexDigest(AESEncrypt(params,key=LINUXAPI_AES_KEY,mode=AES.MODE_ECB))
    }
# endregion



# -*- coding: utf-8 -*-
# PyNCM CLI interface

import logging
from pyncm import (
    DumpSessionAsString,
    GetCurrentSession,
    LoadSessionFromString,
    SetCurrentSession,
    __version__
)
from pyncm.utils.lrcparser import LrcParser
from pyncm.utils.helper import TrackHelper, FuzzyPathHelper, SubstituteWithFullwidth
from pyncm.apis import login, track, playlist, album

from queue import Queue
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from time import sleep
from os.path import join, exists
from os import remove, makedirs

from logging import exception, getLogger, basicConfig
import sys, argparse, re
logger = getLogger('pyncm.main')
# Import checks
OPTIONALS = {"mutagen": False, "tqdm": False, "coloredlogs": False}
OPTIONALS_MISSING_INFO = {
    "mutagen": "无法为下载的音乐添加歌手信息，封面等资源",
    "tqdm": "将不会显示下载进度条",
    "coloredlogs": "日志不会以彩色输出",
}
from importlib.util import find_spec

for import_name in OPTIONALS:
    OPTIONALS[import_name] = find_spec(import_name)
    if not OPTIONALS[import_name]:
        sys.stderr.writelines(
            [f"[WARN] {import_name} 没有安装，{OPTIONALS_MISSING_INFO[import_name]}\n"]
        )

__desc__ = """PyNCM 网易云音乐下载工具 %s""" % __version__

# Key-Value classes
class BaseKeyValueClass:
    def __init__(self, **kw) -> None:
      for k, v in kw.items():
            self.__setattr__(k, v)


class TaskPoolExecutorThread(Thread):
    @staticmethod
    def tag_audio(track: TrackHelper, file: str, cover_img: str = ""):
        if not OPTIONALS["mutagen"]:
            return
        def write_keys(song):
            # Writing metadata
            # Due to different capabilites of containers, only
            # ones that can actually be stored will be written.
            complete_metadata = {
                "title" : [track.TrackName,*(track.TrackAliases or [])],
                "artist" : [*(track.Artists or [])],
                "albumartist" : [*(track.Album.AlbumArtists or [])],
                "album" : [track.AlbumName],
                "tracknumber" :  "%s/%s" % (track.TrackNumber,track.Album.AlbumSongCount),
                "date" : [str(track.Album.AlbumPublishTime)], # TrackPublishTime is not very reliable!
                "copyright": [track.Album.AlbumCompany],
                "discnumber" : [track.CD],
                # These only applies to vorbis tags (e.g. FLAC)                
                "totaltracks" : [str(track.Album.AlbumSongCount)],
                "ncm-id":[str(track.ID)]
            }                    
            for k,v in complete_metadata.items():
                try:
                    song[k] = v
                except:
                    pass
            song.save()

        def mp4():
            from mutagen import easymp4
            from mutagen.mp4 import MP4, MP4Cover

            song = easymp4.EasyMP4(file)
            write_keys(song)
            if exists(cover_img):
                song = MP4(file)
                song["covr"] = [MP4Cover(open(cover_img, "rb").read())]
                song.save()

        def mp3():
            from mutagen.mp3 import EasyMP3,HeaderNotFoundError
            from mutagen.id3 import ID3, APIC
            try:
                song = EasyMP3(file)            
            except HeaderNotFoundError:
                song = EasyMP3()
                song.filename = file
                song.save()
                song = EasyMP3(file)            
            write_keys(song)
            if exists(cover_img):
                song = ID3(file)
                song.update_to_v23()  # better compatibility over v2.4
                song.add(
                    APIC(
                        encoding=3,
                        mime="image/jpeg",
                        type=3,
                        desc="",
                        data=open(cover_img, "rb").read(),
                    )
                )
                song.save(v2_version=3)

        def flac():
            from mutagen.flac import FLAC, Picture
            from mutagen.mp3 import EasyMP3
            song = FLAC(file)            
            write_keys(song)
            if exists(cover_img):
                pic = Picture()
                pic.data = open(cover_img, "rb").read()
                pic.mime = "image/jpeg"
                song.add_picture(pic)   
                song.save()

        def ogg():
            import base64
            from mutagen.flac import Picture
            from mutagen.oggvorbis import OggVorbis

            song = OggVorbis(file)            
            write_keys(song)
            if exists(cover_img):
                pic = Picture()
                pic.data = open(cover_img, "rb").read()
                pic.mime = "image/jpeg"
                song["metadata_block_picture"] = [
                    base64.b64encode(pic.write()).decode("ascii")
                ]
                song.save()

        format = file.split(".")[-1].upper()
        for ext, method in [
            ({"M4A", "M4B", "M4P", "MP4"}, mp4),
            ({"MP3"}, mp3),
            ({"FLAC"}, flac),
            ({"OGG", "OGV"}, ogg),
        ]:
            if format in ext:
                return method() or True
        return False

    def download_by_url(self, url, dest, xfer=False):
        # Downloads generic content
        response = GetCurrentSession().get(url, stream=True)
        length = int(response.headers.get("content-length"))

        with open(dest, "wb") as f:
            for chunk in response.iter_content(128 * 2**10):
                self.xfered += len(chunk)
                if xfer:
                    self.finished_tasks += len(chunk) / length  # task [0,1]
                f.write(chunk)  # write every 128KB read
        return dest

    def __init__(self, *a, max_workers=4, **k):
        super().__init__(*a, **k)
        self.daemon = True
        self.finished_tasks: float = 0
        self.xfered = 0
        self.task_queue = Queue()
        self.max_workers = max_workers

    def run(self):
        def execute(task: BaseKeyValueClass):
            if type(task) == MarkerTask:
                # Mark a finished task w/o execution                
                self.finished_tasks += 1
                return
            if type(task) == TrackDownloadTask:
                try:
                    # Downloding source audio
                    dAudio = track.GetTrackAudioV1(task.audio.id, level=task.audio.level)
                    dAudio = dAudio.get("data", [{"url": ""}])[0]  # Dummy fallback value                
                    assert dAudio["url"], "%s 无法下载，资源不存在" % task.song.Title                
                    logger.info(
                        "开始下载 #%d / %d - %s - %s - %skbps - %s"
                        % (
                            task.index + 1,
                            task.total,
                            task.song.Title,
                            task.song.AlbumName,
                            dAudio["br"] // 1000,
                            dAudio["type"].upper(),
                        )
                    )
                    if not exists(task.audio.dest):
                        makedirs(task.audio.dest)
                    dest_src = self.download_by_url(
                        dAudio["url"],                        
                        task.save_as + "." + dAudio["type"],                    
                        xfer=True,
                    )
                    # Downloading cover
                    dest_cvr = self.download_by_url(
                        task.cover.url, "%s.jpg" % task.save_as
                    )
                    # Downloading & Parsing lyrics
                    dest_lrc = task.save_as
                    lrc = LrcParser()
                    dLyrics = track.GetTrackLyrics(task.lyrics.id)
                    for k in set(dLyrics.keys()) & (
                        {"lrc", "tlyric", "romalrc"} - task.lyrics.lrc_blacklist
                    ):  # Filtering LRCs
                        lrc.LoadLrc(dLyrics[k]["lyric"])
                    lrc_text = lrc.DumpLyrics()
                    if lrc_text:
                        open(dest_lrc, "w", encoding="utf-8").write(lrc_text)
                    # Tagging the audio
                    try:
                        self.tag_audio(task.song, dest_src, dest_cvr)
                    except Exception as e:
                        logger.warning("标签失败 - %s - %s" % (task.song.Title, e))
                    logger.info(
                        "完成下载 #%d / %d - %s" % (task.index + 1, task.total, task.song.Title)
                    )                
                except Exception as e:
                    logger.warning("下载失败 %s - %s" % (task.song.Title, e))
                    task.routine.result_exception(task.song.ID,e,task.song.Title)
                # Cleaning up
                remove(dest_cvr)

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            while True:
                task = self.task_queue.get()
                future = executor.submit(execute, task)
                future.add_done_callback(lambda future: self.task_queue.task_done())


# Subroutines
class Subroutine:
    """Generic subroutine

    Subroutines are `callable`,upon called with `ids`,one
    queues tasks with all given arguments via `put_func` callback
    """
    exceptions = None
    def __init__(self, args, put_func) -> None:
        self.args = args
        self.put = put_func
        self.exceptions = dict()

    def result_exception(self,result_id,exception : Exception,desc=None):
        self.exceptions.setdefault(result_id,list())
        self.exceptions[result_id].append((exception,desc))

    @property
    def has_exceptions(self):
        return len(self.exceptions) > 0
    
class BaseDownloadTask(BaseKeyValueClass):
    id: int
    url: str
    dest: str
    level : str


class LyricsDownloadTask(BaseDownloadTask):
    id: int
    dest: str
    lrc_blacklist: set


class TrackDownloadTask(BaseKeyValueClass):
    song: TrackHelper
    cover: BaseDownloadTask
    lyrics: BaseDownloadTask
    audio: BaseDownloadTask

    index: int
    total: int
    lyrics_exclude: set
    save_as: str

    routine : Subroutine

class MarkerTask(BaseKeyValueClass):
    pass

class Playlist(Subroutine):
    """Base routine for ID-based tasks"""
    def forIds(self, ids):
        dDetails = [track.GetTrackDetail(ids[index:min(len(ids),index+1000)]).get("songs") for index in range(0,len(ids),1000)]
        dDetails = [song for stripe in dDetails for song in stripe]           
        dDetails = sorted(dDetails, key=lambda song: song["id"])
        index = 0
        for index, dDetail in enumerate(dDetails):
            try:
                song = TrackHelper(dDetail)
                output_name=SubstituteWithFullwidth(
                    self.args.output_name.format(**song.template)
                )
                output_folder=self.args.output.format(**{
                    k:SubstituteWithFullwidth(v) for k,v in song.template.items()
                })
                # If audio file already exsists
                # Skip the entire download if `--no_overwrite` is explicitly set                            
                if self.args.no_overwrite:
                    if FuzzyPathHelper(output_folder).exists(
                        output_name,partial_extension_check=True
                    ):
                        logger.warning(
                            "单曲 #%d / %d - %s - %s 已存在，跳过"
                            % (index + 1, len(dDetails), song.Title, song.AlbumName))
                        self.put(MarkerTask())
                        continue                    
                tSong = TrackDownloadTask(
                    index=index,
                    total=len(dDetails),
                    song=song,
                    cover=BaseDownloadTask(
                        id=song.ID, url=song.AlbumCover, dest=output_folder
                    ),
                    audio=BaseDownloadTask(
                        id=song.ID,
                        level=self.args.quality,
                        dest=output_folder,
                    ),
                    lyrics=LyricsDownloadTask(
                        id=song.ID,
                        dest=output_folder,
                        lrc_blacklist=set(self.args.lyric_no),
                    ),
                    save_as=join(output_folder,output_name),
                    routine=self
                )                

                self.put(tSong)

            except Exception as e:                    
                logger.warning(
                    "单曲 #%d / %d - %s - %s 无法下载： %s"
                    % (index + 1, len(dDetails), song.Title, song.AlbumName, e)
                )                
                self.result_exception(song.ID,e,song.Title)
        return index + 1

    def __call__(self, ids):
        queued = 0
        for _id in ids:
            dList = playlist.GetPlaylistInfo(_id)
            logger.info("歌单 ：%s" % dict(dList)["playlist"]["name"])
            queued += self.forIds(
                [tid.get("id") for tid in dict(dList)["playlist"]["trackIds"]]
            )
        return queued

class Album(Playlist):
    def __call__(self, ids):
        queued = 0
        for _id in ids:
            dList = album.GetAlbumInfo(_id)
            logger.info("专辑 ：%s" % dict(dList)["album"]["name"])
            queued += self.forIds([tid["id"] for tid in dList["songs"]])
        return queued

class Song(Playlist):
    def __call__(self, ids):
        return self.forIds(ids)
    
def create_subroutine(sub_type) -> Subroutine:
    """Dynamically creates subroutine callable by string specified"""
    return {"song": Song, "playlist": Playlist, "album": Album}[sub_type]


def parse_sharelink(url):
    """Parses (partial) URLs for NE resources and determines its ID and type

    e.g.
        31140560 (plain song id)
        https://greats3an.github.io/pyncmd/?trackId=1818064296 (pyncmd)
        分享Ali Edwards的单曲《Devil Trigger》: http://music.163.com/song/1353163404/?userid=6483697162 (来自@网易云音乐) (mobile app)
        "分享mos9527创建的歌单「東方 PC」: http://music.163.com/playlist?id=72897851187" (desktop app)
    """
    rurl = re.findall("(?:http|https):\/\/.*", url)
    if rurl:
        url = rurl[0]  # Use first URL found. Otherwise use value given as is.
    numerics = re.findall("\d{4,}", url)
    assert numerics, "未在链接中找到任何 ID"
    ids = numerics[:1]  # Only pick the first match
    table = {
        "song": ["trackId", "song"],
        "playlist": ["playlist"],
        "album": ["album"],
    }
    rtype = "song"  # Defaults to songs (tracks)
    for rtype_, rkeyword in table.items():
        for kw in rkeyword:
            if kw in url:
                rtype = rtype_
                break  # Match type by keyword
    return rtype, ids


PLACEHOLDER_URL = "00000"


def parse_args():
    """Setting up __main__ argparser"""
    parser = argparse.ArgumentParser(
        description=__desc__, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "url", metavar="链接", help="网易云音乐分享链接", nargs="*", default=PLACEHOLDER_URL
    )
    group = parser.add_argument_group("下载")
    group.add_argument(
        "--max-workers", "-m", metavar="最多同时下载任务数", default=4, type=int
    )
    group.add_argument(
        "--output-name",
        "--template",
        "-t",
        metavar="保存文件名模板",
        help=r"""保存文件名模板
    参数：    
        id     - 网易云音乐资源 ID
        year   - 出版年份
        no     - 专辑中编号
        album  - 专辑标题
        track  - 单曲标题        
        title  - 完整标题
        artists- 艺术家名
    例：
        {track} - {artists} 等效于 {title}""",
        default=r"{title}",
    )
    group.add_argument("-o","--output", metavar="输出", default=".", help=r"""输出文件夹
    注：该参数也可使用模板，格式同 保存文件名模板"""
    )
    group.add_argument(
        "--quality",
        metavar="音质",
        choices=['standard','exhigh','lossless','hi-res'],
        help=r"""音频音质（高音质需要 CVIP）
    参数：
        hi-res  - Hi-Res
        lossless- “无损”
        exhigh  - 较高
        standard- 标准""",
        default="standard",
    )
    group.add_argument("--no-overwrite", action="store_true", help="不重复下载已经存在的音频文件")
    group = parser.add_argument_group("歌词")
    group.add_argument(
        "--lyric-no",
        metavar="跳过歌词",
        help=r"""跳过某些歌词类型的合并
    参数：
        lrc    - 源语言歌词
        tlyric - 翻译后歌词
        romalrc- 罗马音歌词
    例：
        --lyric-no tlyric --lyric-no romalrc 将只下载源语言歌词""",
        choices=["lrc", "tlyric", "romalrc"],
        default="",
        nargs="+",
    )
    group = parser.add_argument_group("登陆")
    group.add_argument("--phone", metavar="手机", default="", help="网易账户手机号")
    group.add_argument("--pwd", "--password", metavar="密码", default="", help="网易账户密码")
    group.add_argument("--save", metavar="[保存到]", default="", help="写本次登录信息于文件")
    group.add_argument(
        "--load", metavar="[保存的登陆信息文件]", default="", help="从文件读取登录信息供本次登陆使用"
    )
    group.add_argument("--http", action="store_true", help="优先使用 HTTP，不保证不被升级")
    group.add_argument("--log-level", help="日志等级", default="INFO")

    args = parser.parse_args()
    
    try:
        return args , [parse_sharelink(url) for url in args.url]
    except AssertionError:
        if args.url == PLACEHOLDER_URL:
            sys.argv.append("-h")  # If using placeholder, no argument is really passed
            return __main__()  # In which case, print help and exit
        assert args.save, "无效分享链接 %s" % ' '.join(args.url) # Allow invalid links for this one            
        return args , []

def __main__():    
    args , tasks = parse_args()    
    log_stream = sys.stdout
    # Getting tqdm & logger to work nicely together
    if OPTIONALS["tqdm"]:
        from tqdm.std import tqdm as tqdm_c

        class SemaphoreStdout:
            @staticmethod
            def write(__s):
                # Blocks tqdm's output until write on this stream is done
                # Solves cases where progress bars gets re-rendered when logs
                # spews out too fast
                with tqdm_c.external_write_mode(file=sys.stdout, nolock=False):
                    return sys.stdout.write(__s)

        log_stream = SemaphoreStdout
    if OPTIONALS["coloredlogs"]:
        import coloredlogs

        coloredlogs.install(
            level=args.log_level,
            fmt="%(asctime)s %(name)s [%(levelname).4s] %(message)s",
            stream=log_stream,
            isatty=True,
        )
    basicConfig(
        level=args.log_level, format="[%(levelname).4s] %(name)s %(message)s", stream=log_stream
    )
    if args.load:
        logger.info("读取登录信息 : %s" % args.load)
        SetCurrentSession(LoadSessionFromString(open(args.load).read()))
    if args.http:
        GetCurrentSession().force_http = True
        logger.warning("优先使用 HTTP")
    if args.phone and args.pwd:
        login.LoginViaCellphone(args.phone, args.pwd)
        logger.info(
            "账号 ：%s (VIP %s)"
            % (GetCurrentSession().nickname, GetCurrentSession().vipType)
        )
    if args.save:
        logger.info("保存登陆信息于 : %s" % args.save)
        open(args.save, "w").write(DumpSessionAsString(GetCurrentSession()))
        return 0
    if not GetCurrentSession().logged_in:
        login.LoginViaAnonymousAccount()
        logger.info("以匿名身份登陆成功，UID: %s" % GetCurrentSession().uid)
    executor = TaskPoolExecutorThread(max_workers=args.max_workers)
    executor.daemon = True
    executor.start()

    def enqueue_task(task):
        executor.task_queue.put(task)

    total_queued = 0
    subroutines = []
    for rtype,ids in tasks:
        task_desc = "ID: %s 类型: %s" % (ids[0],{'album':'专辑','playlist':'歌单®','song':'单曲'}[rtype])
        logger.info("处理任务 %s" % task_desc) 
        subroutine = create_subroutine(rtype)(args, enqueue_task)
        total_queued += subroutine(ids)  # Enqueue tasks
        subroutines.append((subroutine,task_desc))

    if OPTIONALS["tqdm"]:
        import tqdm

        _tqdm = tqdm.tqdm(
            bar_format="{desc}: {percentage:.1f}%|{bar}| {n:.2f}/{total_fmt} {elapsed}<{remaining}"
        )
        _tqdm.total = total_queued

        def report():
            _tqdm.desc = _tqdm.format_sizeof(executor.xfered, suffix="B", divisor=1024)
            _tqdm.update(min(executor.finished_tasks, total_queued) - _tqdm.n)
            return True

    else:

        def report():
            sys.stderr.write(
                f"下载中 : {executor.finished_tasks:.1f} / {total_queued} ({(executor.finished_tasks * 100 / total_queued):.1f} %,{executor.xfered >> 20} MB)               \r"
            )
            return True

    while executor.task_queue.unfinished_tasks >= 0:
        try:
            report() and sleep(0.5)
            if executor.task_queue.unfinished_tasks == 0:
                break
        except KeyboardInterrupt:
            break
    
    # Check final results
    # Thought: Maybe we should automatically retry these afterwards?
    failed_ids = dict()
    for routine,desc in subroutines:
        routine : Subroutine
        if routine.has_exceptions:
            logger.error('%s - 下载未完成' % desc)
            for exception_id , exceptions in routine.exceptions.items():
                failed_ids[exception_id] = True
                for exception in exceptions:
                    exception_obj, desc = exception
                    logger.warning('下载出错 ID: %s - %s%s' % (exception_id,exception_obj,' (%s)' % desc if desc else ''))
    if failed_ids:
        logger.error('你可以将下载失败的 ID 作为参数以再次下载')
        logger.error('所有失败的任务 ID: %s' % ' '.join([str(i) for i in failed_ids.keys()]))
    report()
    logger.info(f'任务完成率 {(executor.finished_tasks * 100 / total_queued):.1f}%')
    return 0


if __name__ == "__main__":
    sys.exit(__main__())
